//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                          //
//                                                    TM                                                    //
//                                        Intelligrill   Feather Server                                     //
//                                                                                                          //
//                                                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

#define PERFBOARD   1

//
//             TM
// Intelligrill   Feather Server.
//
// Copyright (c) 2017 by Zumwalt Properties, LLC., All Rights Reserved.
//
//             TM
// Intelligrill   is a registered trademark of Zumwalt Properties, LLC., All Rights Reserved
//

//
//             TM
// Intelligrill   Feather Parts List.
//
//  1) Adafruit "Feather Huzzah ESP8266".
//  2) Adafruit "FeatherWing OLED - 128x32 OLED Add-on For All Feather Boards".
//  3) Maverick "ET-72 Temperature Probe".
//  4) 2.5mm audio connector, panel mount.
//  5) 22k ohm 1% 1/8 watt resistor.
//  6) 680 ohm 1% 1/8 watt resistor.
//  7) Analog Devices "1vdc reference source" (ADR510).
//

//
//             TM
// Intelligrill   Feather Setup.
//
//     With the settings I have provided in Intelligrill, during the first start up, Intelligrill
//  attempts to connect to a wifi network with an ssid "your_ssid" and password "Intelligrill".
//  At the same time, Intelligrill creates an "access point" network with the ssid "Intelligrill"
//  and password "Intelligrill".  In order to provide Intellgrill with access to your wifi network,
//  you will need to connect to the Intelligrill access point network in order to alter the wifi
//  settings to allow it to connect to your network. The steps involved in doing so follow:
//
//  1) Download the Intelligrill software to your assembled Intelligrill and wait for it to show
//     the "IP Address" page on the Intelligrill OLED display.  At this point, unless your wifi
//     router ssid is "your_ssid" and the password is "Intelligrill", the ip address shown on this
//     page should read "0.0.0.0".  This is normal as you have yet to program the ssid and password
//     of your wifi network into Intelligrill.  For security purposes, access to the Intelligrill
//     setup web page is only allowed when the Intelligrill is displaying the "IP Address" page, so
//     leave the Intelligrill on the "IP Address" page until setup is completed.
//
//  2) On a wifi enabled web browser device, go to the wifi settings and select the "Intelligrill"
//     network.
//
//  3) When the wifi settings ask for a password for the Intelligrill network, enter "Intelligrill".
//
//  4) Once the network connects (this can take some time for reasons I have yet to determine), start
//     a web browser on your wifi enabled web browser device and enter "192.168.20.20/setup" into the
//     url field of the web browser.
//
//  5) The Intelligrill Setup page should appear on your browser.  If your Intelligrill was not on
//     the "IP Address" page, then a friendly reminder will appear in your web browser reminding
//     you as such.
//
//  6) If you wish to alter the Intelligrill ssid (for example, if you will be using more than one
//     Intelligrill at a time, they will require different ssids), enter the desired Intelligrill
//     ssid in the "Intelligrill ssid:" box.  Since I have a static IP address from my internet
//     service provider, I reserve fixed IP addresses for my Intelligrills on my router for each
//     Intelligrill I use, then enable port forwarding and assign a unique port number for each
//     Intelligrill, thus I set the ssid of each intelligrill to be "Intelligrill" + port number
//     (e.g. "Intelligrill2205").  Using this method, I can monitor each Intelligrill from anywhere
//     I have internet access.
//
//  7) Enter the ssid of your wifi router in the "Wifi ssid:" box.
//
//  8) Enter the password of your wifi router in the "Intelligrill & Wifi password:" box.  Your wifi router
//     password will also become the password for the "Intelligrill" access point for any future access to
//     the Intelligrill access point.
//
//  9) Click "Save".  If you happen to have the Arduino IDE connected and serial monitor open, you should
//     see the "Intelligrill data written to storage":' message followed by the Intelligrill ssid, Wifi
//     ssid, and password you entered.  This is a good thing.
//
// 10) On your wifi enabled web browser device, logon to your wifi network and then in your wifi settings
//     "forget" the Intelligrill network (since the password has now been changed, connecting to this network
//     is no longer possible).
//
// 11) Reboot the Intelligrill (press the reset button).
//
// 12) On the Intelligrill display, after the title and copy right pages cycle, the IP Address page should
//     now display an ip address provided by your network router as anything other than "0.0.0.0".  Typically,
//     the display will show something like "192.168.X.X". where X indicates values provided by your router.
//     And again if you happen to have the Arduino IDE connected and serial monitor open, you should see the
//     "Intelligrill data recovered from storage:" message followed by the Intelligrill ssid, WiFi ssid, and
//     password you entered.  This is a very good thing.
//
// 13) Enter the IP address that appears on the Intelligrill display into the web browser url window on your
//     wifi enabled web browser device, and when the Intelligrill page appears, you are cooking!
//
// 14) Caveats:
//
//     a) The esp8266 does not provide a mechanism to write your ssids and password into flash memory.  As
//        such, Intelligrill Feather Server writes these values into the real time clock memory.  If your
//        Intelligrill battery every becomes fully depleted, then you will have to repeat the above setup
//        process to give your Intelligrill access to your wifi network.  I recommend putting Intelligrill
//        in "power down" mode (press and hold the center button until the "Good Night !" message appears)
//        when not in use and keep it connected to a usb power source in order to keep the battery fully
//        charged.  And for long term grilling / smoking, either leave your Intelligrill plugged into mains
//        AC, or in the event your location does not have access to mains AC, simply use an external cellphone
//        style battery extender or other usb compatible battery source with a usb to micro usb cable connected
//        between the battery source and micro usb port.
//
// 15) If at any time you believe the settings process has become hoplessly lost, remove both the usb and
//     battery connections to completely power down the Intelligrill, wait a minute or so, then recconect
//     power and repeat the setup process from the beginning.
//

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Library includes.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include                        <Adafruit_FeatherOLED.h>
#include                        <Adafruit_FeatherOLED_WiFi.h>
#include                        <ESP8266WebServer.h>
extern "C" 
{
                                // This include is for access to the ESP8266 real time clock (rtc) ram memory
                                // used as storage for Intelligrill ssids and password during power down.
                                
                                #include "user_interface.h"
}
#include                        <FS.h>

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Constants.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

  // Analog input count.

      #define                   AD_COUNT_MAX      1023.0            // Maximum analog input count.
      
  // Buttons on Oled.
  
      #define                   BUTTON_A          0                 // Button A pin.
      #define                   BUTTON_B          16                // Button B pin.
      #define                   BUTTON_C          2                 // Button C pin.

  // Counts averaging array size.
  
      #define                   COUNTS_ARRAY_SIZE    20
      
  // Oled display.

      // Intelligrill display modes.
      
      enum                      DISPLAY_MODES
      {
                                DISPLAY_MODE_IP,                    // Display the ip address of the feather board.
                                DISPLAY_MODE_TEMPERATURE,           // Display the current temperature.
                                DISPLAY_MODE_TEMPERATURE_DESIRED,   // Display the user input desired temperature.
                                DISPLAY_MODE_TIME_PREDICTED_TO_GO,  // Display the estimated time to go.
                                DISPLAY_MODE_TIME_RUNNING,          // Display the time running.
                                DISPLAY_MODE_COUNT
      };

      // Dimensions.
      
      #define                   DISPLAY_HEIGHT    32                // Height of afOled display in pixels.
      #define                   DISPLAY_WIDTH     128               // Width of afOled display in pixels.

      // Font dimensions (estimated).
      
      #define                   FONT_ONE_HEIGHT   8                 // Height of font one character.
      #define                   FONT_ONE_WIDTH    6                 // Width of font one character.
      #define                   FONT_TWO_HEIGHT   14                // Height of font two character.
      #define                   FONT_TWO_WIDTH    12                // Width of font two character.
  
  // Real time clock ram.

      // Start of rtc memory.
      
      #define                       RTC_MEMORY_START  64                // Start of integer real time clock ram.

      // Structure used to store and retrieve ssid and password from rtc memory.  Note the inclusion of a crc
      // value used to verify the intergrity of the data stored.  Note also the rtc memory is in blocks of 4
      // bytes, thus the structure element sizes are in 4 byte multiples.
      
      typedef                       struct                              // Structure to store ssid, password and crc in real time clock ram.
                                    {
                                      uint8     chSsid[36];             // Maximum ssid length is 32 bytes, limited by Intelligrill_Feather_Setup.h.
                                      uint8     chSsidAp[36];           // Maximum ssid length is 32 bytes, limited by Intelligrill_Feather_Setup.h.
                                      uint8     chPassword[68];         // Maximum password length is 64 bytes, limited by Intelligrill_Feather_Setup.h. 
                                      uint32    nCrc;   
                                    } rtcStorage;
    
  // Timing.
  
  #define                       DELAY_MILLISECONDS  2               // Main loop delay.

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Variables.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

      // Copyright.
      
      const char*               chCopyright =
                                "Intelligrill (TM) Feather Copyright 2017 by Zumwalt Properties, LLC.  All Rights Reserved";
                                
      // Access point.
      
      String                    sSsidIntelligrill = "Intelligrill";
      IPAddress                 IpAddressIntelligrill(192,168,20,20);
      IPAddress                 IpGatewayIntelligrill(192,168,20,1);
      IPAddress                 IpSubnetIntelligrill(255,255,255,0);

      // Analog.
      //
      // Notes:
      //
      //  The temperature table was designed for the original pic analog input range of 0 through 1023 counts.
      // The Feather ESP8266 range is less than 0 through 1023 counts (approximately 3 through 978) thus this
      // compensation is to offset and expand the Feather ESP8266 range to the pic range.  The offset and scale
      // were determined by grounding the analog input then placing 1.00 volts on the analog input, observing
      // the counts for each, then determining the offset and scale from those values.
      //

#if PERFBOARD == 1

      double                    dCountsOffset = 46.0;
      double                    dCountsScale = 1.0556701;
    
#else

      double                    dCountsOffset = 52.0;
      double                    dCountsScale = 1.0556701;
      
#endif

      int                       nCounts[COUNTS_ARRAY_SIZE];
      int                       nCountsPointer = 0;

      // Display.
      
      int                       nDisplayMode = DISPLAY_MODE_IP;

      // Editing.
      
      bool                      bEdit = false;
      int                       nEditIncrement = 0;

      // Oled.
      
      Adafruit_FeatherOLED_WiFi afOled = Adafruit_FeatherOLED_WiFi();

      // Port number.
      
      const int                 nPort = 80;

      // Real time clock memory image.

      rtcStorage                rtcRamImage;
      
      // Run / stop.
      
      bool                      bRunning = false;

      // Seconds off / on.
      
      bool                      bSecondsDisplay = false;
      
      // Server.
      
      ESP8266WebServer          esp8266Server(nPort);

      // ssid and password.
      
      String                    sSsid = "your_ssid";
      String                    sPassword = "Intelligrill";

      // Temperature.
      
      int                       nTemperatureCurrent = 0;
      int                       nTemperatureDesired = 165;
      double                    dTemperatureF = 0.0;
      int                       nTemperatureHighest = -999;
      int                       nTemperatureLowest = 999;
      const int                 nTemperatureTable[(int)(AD_COUNT_MAX + 1.0)] =
                                {
                                  0x270f, 0x270f, 0x270f, 0x2706, 0x2451, 0x221c, 0x206e, 0x1f16, 0x1dfa, 0x1d09, 
                                  0x1c39, 0x1b82, 0x1ae0, 0x1a4e, 0x19ca, 0x1952, 0x18e3, 0x187d, 0x181f, 0x17c6, 
                                  0x1774, 0x1726, 0x16dd, 0x1698, 0x1657, 0x1619, 0x15de, 0x15a6, 0x1571, 0x153d, 
                                  0x150c, 0x14dd, 0x14b0, 0x1484, 0x145a, 0x1432, 0x140a, 0x13e5, 0x13c0, 0x139d, 
                                  0x137a, 0x1359, 0x1339, 0x1319, 0x12fb, 0x12dd, 0x12c0, 0x12a4, 0x1289, 0x126e, 
                                  0x1254, 0x123a, 0x1222, 0x1209, 0x11f1, 0x11da, 0x11c4, 0x11ad, 0x1197, 0x1182, 
                                  0x116d, 0x1159, 0x1145, 0x1131, 0x111d, 0x110a, 0x10f8, 0x10e5, 0x10d3, 0x10c2, 
                                  0x10b0, 0x109f, 0x108e, 0x107e, 0x106d, 0x105d, 0x104d, 0x103e, 0x102f, 0x101f, 
                                  0x1011, 0x1002, 0x0ff3, 0x0fe5, 0x0fd7, 0x0fc9, 0x0fbc, 0x0fae, 0x0fa1, 0x0f94, 
                                  0x0f87, 0x0f7a, 0x0f6d, 0x0f61, 0x0f54, 0x0f48, 0x0f3c, 0x0f30, 0x0f25, 0x0f19, 
                                  0x0f0d, 0x0f02, 0x0ef7, 0x0eec, 0x0ee1, 0x0ed6, 0x0ecb, 0x0ec0, 0x0eb6, 0x0eab, 
                                  0x0ea1, 0x0e97, 0x0e8d, 0x0e83, 0x0e79, 0x0e6f, 0x0e65, 0x0e5c, 0x0e52, 0x0e49, 
                                  0x0e3f, 0x0e36, 0x0e2d, 0x0e24, 0x0e1b, 0x0e12, 0x0e09, 0x0e00, 0x0df8, 0x0def, 
                                  0x0de6, 0x0dde, 0x0dd6, 0x0dcd, 0x0dc5, 0x0dbd, 0x0db5, 0x0dac, 0x0da4, 0x0d9d, 
                                  0x0d95, 0x0d8d, 0x0d85, 0x0d7d, 0x0d76, 0x0d6e, 0x0d67, 0x0d5f, 0x0d58, 0x0d50, 
                                  0x0d49, 0x0d42, 0x0d3a, 0x0d33, 0x0d2c, 0x0d25, 0x0d1e, 0x0d17, 0x0d10, 0x0d09, 
                                  0x0d03, 0x0cfc, 0x0cf5, 0x0cee, 0x0ce8, 0x0ce1, 0x0cda, 0x0cd4, 0x0ccd, 0x0cc7, 
                                  0x0cc1, 0x0cba, 0x0cb4, 0x0cae, 0x0ca7, 0x0ca1, 0x0c9b, 0x0c95, 0x0c8f, 0x0c89, 
                                  0x0c83, 0x0c7d, 0x0c77, 0x0c71, 0x0c6b, 0x0c65, 0x0c5f, 0x0c59, 0x0c54, 0x0c4e, 
                                  0x0c48, 0x0c42, 0x0c3d, 0x0c37, 0x0c32, 0x0c2c, 0x0c27, 0x0c21, 0x0c1c, 0x0c16, 
                                  0x0c11, 0x0c0b, 0x0c06, 0x0c01, 0x0bfb, 0x0bf6, 0x0bf1, 0x0bec, 0x0be6, 0x0be1, 
                                  0x0bdc, 0x0bd7, 0x0bd2, 0x0bcd, 0x0bc8, 0x0bc3, 0x0bbe, 0x0bb9, 0x0bb4, 0x0baf, 
                                  0x0baa, 0x0ba5, 0x0ba0, 0x0b9b, 0x0b96, 0x0b91, 0x0b8d, 0x0b88, 0x0b83, 0x0b7e, 
                                  0x0b7a, 0x0b75, 0x0b70, 0x0b6c, 0x0b67, 0x0b62, 0x0b5e, 0x0b59, 0x0b55, 0x0b50, 
                                  0x0b4b, 0x0b47, 0x0b42, 0x0b3e, 0x0b3a, 0x0b35, 0x0b31, 0x0b2c, 0x0b28, 0x0b24, 
                                  0x0b1f, 0x0b1b, 0x0b17, 0x0b12, 0x0b0e, 0x0b0a, 0x0b05, 0x0b01, 0x0afd, 0x0af9, 
                                  0x0af4, 0x0af0, 0x0aec, 0x0ae8, 0x0ae4, 0x0ae0, 0x0adc, 0x0ad7, 0x0ad3, 0x0acf, 
                                  0x0acb, 0x0ac7, 0x0ac3, 0x0abf, 0x0abb, 0x0ab7, 0x0ab3, 0x0aaf, 0x0aab, 0x0aa7, 
                                  0x0aa3, 0x0aa0, 0x0a9c, 0x0a98, 0x0a94, 0x0a90, 0x0a8c, 0x0a88, 0x0a84, 0x0a81, 
                                  0x0a7d, 0x0a79, 0x0a75, 0x0a71, 0x0a6e, 0x0a6a, 0x0a66, 0x0a63, 0x0a5f, 0x0a5b, 
                                  0x0a57, 0x0a54, 0x0a50, 0x0a4c, 0x0a49, 0x0a45, 0x0a41, 0x0a3e, 0x0a3a, 0x0a37, 
                                  0x0a33, 0x0a2f, 0x0a2c, 0x0a28, 0x0a25, 0x0a21, 0x0a1e, 0x0a1a, 0x0a17, 0x0a13, 
                                  0x0a10, 0x0a0c, 0x0a09, 0x0a05, 0x0a02, 0x09fe, 0x09fb, 0x09f7, 0x09f4, 0x09f0, 
                                  0x09ed, 0x09ea, 0x09e6, 0x09e3, 0x09df, 0x09dc, 0x09d9, 0x09d5, 0x09d2, 0x09cf, 
                                  0x09cb, 0x09c8, 0x09c5, 0x09c1, 0x09be, 0x09bb, 0x09b7, 0x09b4, 0x09b1, 0x09ae, 
                                  0x09aa, 0x09a7, 0x09a4, 0x09a1, 0x099d, 0x099a, 0x0997, 0x0994, 0x0991, 0x098d, 
                                  0x098a, 0x0987, 0x0984, 0x0981, 0x097d, 0x097a, 0x0977, 0x0974, 0x0971, 0x096e, 
                                  0x096b, 0x0967, 0x0964, 0x0961, 0x095e, 0x095b, 0x0958, 0x0955, 0x0952, 0x094f, 
                                  0x094c, 0x0949, 0x0945, 0x0942, 0x093f, 0x093c, 0x0939, 0x0936, 0x0933, 0x0930, 
                                  0x092d, 0x092a, 0x0927, 0x0924, 0x0921, 0x091e, 0x091b, 0x0918, 0x0915, 0x0912, 
                                  0x090f, 0x090c, 0x0909, 0x0906, 0x0904, 0x0901, 0x08fe, 0x08fb, 0x08f8, 0x08f5, 
                                  0x08f2, 0x08ef, 0x08ec, 0x08e9, 0x08e6, 0x08e3, 0x08e1, 0x08de, 0x08db, 0x08d8, 
                                  0x08d5, 0x08d2, 0x08cf, 0x08cc, 0x08ca, 0x08c7, 0x08c4, 0x08c1, 0x08be, 0x08bb, 
                                  0x08b9, 0x08b6, 0x08b3, 0x08b0, 0x08ad, 0x08aa, 0x08a8, 0x08a5, 0x08a2, 0x089f, 
                                  0x089c, 0x089a, 0x0897, 0x0894, 0x0891, 0x088e, 0x088c, 0x0889, 0x0886, 0x0883, 
                                  0x0881, 0x087e, 0x087b, 0x0878, 0x0876, 0x0873, 0x0870, 0x086d, 0x086b, 0x0868, 
                                  0x0865, 0x0862, 0x0860, 0x085d, 0x085a, 0x0858, 0x0855, 0x0852, 0x084f, 0x084d, 
                                  0x084a, 0x0847, 0x0845, 0x0842, 0x083f, 0x083d, 0x083a, 0x0837, 0x0835, 0x0832, 
                                  0x082f, 0x082d, 0x082a, 0x0827, 0x0825, 0x0822, 0x081f, 0x081d, 0x081a, 0x0817, 
                                  0x0815, 0x0812, 0x080f, 0x080d, 0x080a, 0x0807, 0x0805, 0x0802, 0x07ff, 0x07fd, 
                                  0x07fa, 0x07f8, 0x07f5, 0x07f2, 0x07f0, 0x07ed, 0x07ea, 0x07e8, 0x07e5, 0x07e3, 
                                  0x07e0, 0x07dd, 0x07db, 0x07d8, 0x07d6, 0x07d3, 0x07d0, 0x07ce, 0x07cb, 0x07c9, 
                                  0x07c6, 0x07c3, 0x07c1, 0x07be, 0x07bc, 0x07b9, 0x07b6, 0x07b4, 0x07b1, 0x07af, 
                                  0x07ac, 0x07aa, 0x07a7, 0x07a4, 0x07a2, 0x079f, 0x079d, 0x079a, 0x0798, 0x0795, 
                                  0x0792, 0x0790, 0x078d, 0x078b, 0x0788, 0x0786, 0x0783, 0x0781, 0x077e, 0x077b, 
                                  0x0779, 0x0776, 0x0774, 0x0771, 0x076f, 0x076c, 0x076a, 0x0767, 0x0764, 0x0762, 
                                  0x075f, 0x075d, 0x075a, 0x0758, 0x0755, 0x0753, 0x0750, 0x074e, 0x074b, 0x0749, 
                                  0x0746, 0x0743, 0x0741, 0x073e, 0x073c, 0x0739, 0x0737, 0x0734, 0x0732, 0x072f, 
                                  0x072d, 0x072a, 0x0728, 0x0725, 0x0723, 0x0720, 0x071d, 0x071b, 0x0718, 0x0716, 
                                  0x0713, 0x0711, 0x070e, 0x070c, 0x0709, 0x0707, 0x0704, 0x0702, 0x06ff, 0x06fd, 
                                  0x06fa, 0x06f8, 0x06f5, 0x06f2, 0x06f0, 0x06ed, 0x06eb, 0x06e8, 0x06e6, 0x06e3, 
                                  0x06e1, 0x06de, 0x06dc, 0x06d9, 0x06d7, 0x06d4, 0x06d2, 0x06cf, 0x06cc, 0x06ca, 
                                  0x06c7, 0x06c5, 0x06c2, 0x06c0, 0x06bd, 0x06bb, 0x06b8, 0x06b6, 0x06b3, 0x06b1, 
                                  0x06ae, 0x06ac, 0x06a9, 0x06a6, 0x06a4, 0x06a1, 0x069f, 0x069c, 0x069a, 0x0697, 
                                  0x0695, 0x0692, 0x0690, 0x068d, 0x068a, 0x0688, 0x0685, 0x0683, 0x0680, 0x067e, 
                                  0x067b, 0x0679, 0x0676, 0x0673, 0x0671, 0x066e, 0x066c, 0x0669, 0x0667, 0x0664, 
                                  0x0661, 0x065f, 0x065c, 0x065a, 0x0657, 0x0655, 0x0652, 0x064f, 0x064d, 0x064a, 
                                  0x0648, 0x0645, 0x0642, 0x0640, 0x063d, 0x063b, 0x0638, 0x0636, 0x0633, 0x0630, 
                                  0x062e, 0x062b, 0x0629, 0x0626, 0x0623, 0x0621, 0x061e, 0x061b, 0x0619, 0x0616, 
                                  0x0614, 0x0611, 0x060e, 0x060c, 0x0609, 0x0606, 0x0604, 0x0601, 0x05ff, 0x05fc, 
                                  0x05f9, 0x05f7, 0x05f4, 0x05f1, 0x05ef, 0x05ec, 0x05e9, 0x05e7, 0x05e4, 0x05e1, 
                                  0x05df, 0x05dc, 0x05d9, 0x05d7, 0x05d4, 0x05d1, 0x05cf, 0x05cc, 0x05c9, 0x05c7, 
                                  0x05c4, 0x05c1, 0x05be, 0x05bc, 0x05b9, 0x05b6, 0x05b4, 0x05b1, 0x05ae, 0x05ab, 
                                  0x05a9, 0x05a6, 0x05a3, 0x05a0, 0x059e, 0x059b, 0x0598, 0x0595, 0x0593, 0x0590, 
                                  0x058d, 0x058a, 0x0588, 0x0585, 0x0582, 0x057f, 0x057c, 0x057a, 0x0577, 0x0574, 
                                  0x0571, 0x056e, 0x056c, 0x0569, 0x0566, 0x0563, 0x0560, 0x055d, 0x055b, 0x0558, 
                                  0x0555, 0x0552, 0x054f, 0x054c, 0x0549, 0x0547, 0x0544, 0x0541, 0x053e, 0x053b, 
                                  0x0538, 0x0535, 0x0532, 0x052f, 0x052c, 0x052a, 0x0527, 0x0524, 0x0521, 0x051e, 
                                  0x051b, 0x0518, 0x0515, 0x0512, 0x050f, 0x050c, 0x0509, 0x0506, 0x0503, 0x0500, 
                                  0x04fd, 0x04fa, 0x04f7, 0x04f4, 0x04f1, 0x04ee, 0x04eb, 0x04e8, 0x04e4, 0x04e1, 
                                  0x04de, 0x04db, 0x04d8, 0x04d5, 0x04d2, 0x04cf, 0x04cc, 0x04c8, 0x04c5, 0x04c2, 
                                  0x04bf, 0x04bc, 0x04b9, 0x04b5, 0x04b2, 0x04af, 0x04ac, 0x04a9, 0x04a5, 0x04a2, 
                                  0x049f, 0x049c, 0x0498, 0x0495, 0x0492, 0x048f, 0x048b, 0x0488, 0x0485, 0x0481, 
                                  0x047e, 0x047b, 0x0477, 0x0474, 0x0470, 0x046d, 0x046a, 0x0466, 0x0463, 0x045f, 
                                  0x045c, 0x0458, 0x0455, 0x0452, 0x044e, 0x044b, 0x0447, 0x0443, 0x0440, 0x043c, 
                                  0x0439, 0x0435, 0x0432, 0x042e, 0x042a, 0x0427, 0x0423, 0x0420, 0x041c, 0x0418, 
                                  0x0414, 0x0411, 0x040d, 0x0409, 0x0406, 0x0402, 0x03fe, 0x03fa, 0x03f6, 0x03f3, 
                                  0x03ef, 0x03eb, 0x03e7, 0x03e3, 0x03df, 0x03db, 0x03d7, 0x03d3, 0x03cf, 0x03cb, 
                                  0x03c7, 0x03c3, 0x03bf, 0x03bb, 0x03b7, 0x03b3, 0x03af, 0x03ab, 0x03a6, 0x03a2, 
                                  0x039e, 0x039a, 0x0395, 0x0391, 0x038d, 0x0389, 0x0384, 0x0380, 0x037b, 0x0377, 
                                  0x0373, 0x036e, 0x036a, 0x0365, 0x0361, 0x035c, 0x0357, 0x0353, 0x034e, 0x0349, 
                                  0x0345, 0x0340, 0x033b, 0x0336, 0x0331, 0x032d, 0x0328, 0x0323, 0x031e, 0x0319, 
                                  0x0314, 0x030f, 0x030a, 0x0304, 0x02ff, 0x02fa, 0x02f5, 0x02ef, 0x02ea, 0x02e5, 
                                  0x02df, 0x02da, 0x02d4, 0x02cf, 0x02c9, 0x02c3, 0x02be, 0x02b8, 0x02b2, 0x02ac, 
                                  0x02a7, 0x02a1, 0x029b, 0x0294, 0x028e, 0x0288, 0x0282, 0x027c, 0x0275, 0x026f, 
                                  0x0268, 0x0262, 0x025b, 0x0254, 0x024e, 0x0247, 0x0240, 0x0239, 0x0232, 0x022b, 
                                  0x0223, 0x021c, 0x0215, 0x020d, 0x0205, 0x01fe, 0x01f6, 0x01ee, 0x01e6, 0x01de, 
                                  0x01d5, 0x01cd, 0x01c4, 0x01bc, 0x01b3, 0x01aa, 0x01a1, 0x0197, 0x018e, 0x0184, 
                                  0x017b, 0x0171, 0x0166, 0x015c, 0x0152, 0x0147, 0x013c, 0x0131, 0x0125, 0x0119, 
                                  0x010d, 0x0101, 0x00f4, 0x00e7, 0x00da, 0x00cc, 0x00be, 0x00af, 0x00a0, 0x0091, 
                                  0x0081, 0x0070, 0x005f, 0x004d, 0x003a, 0x0027, 0x0012, 0x0000, 0x0000, 0x0000, 
                                  0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 
                                  0x0000, 0x0000, 0x0000, 0x0000
                                };

      // Timing.
      
      double                    dMillisecondsLast = 0.0;
      double                    dMillisecondsRunning = 0.0;
      double                    dTimePredicted = -1.0;
      double                    dTimePredictedToGo = -1.0;
      double                    dTimeRunning = 0.0;
      int                       nTurnOffDelay = 0;

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
//                                                  TM                                                   
//                                       Intelligrill   Feather Client Include                                  
//
// 
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

      #include                  "Intelligrill_Feather_Client.h"

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
//                                                  TM                                                   
//                                       Intelligrill   Feather Setup Include                                  
//
// 
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

      #include                  "Intelligrill_Feather_Client_Setup.h"

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
// Setup.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

void setup()
{
  // Serial.
  //
  // Serial is only necessary for debugging.
  
  Serial.begin(115200);
  while(!Serial){};

  Serial.println(String("\n\n\n\n" + String(chCopyright) + "\n"));
   
  // OLED display.
  //
  // The OLED display in this design is an "Adafruit FeatherWing OLED - 128x32 OLED".
  
      // Initialize the oled.
      
      afOled.init();
      afOled.clearDisplay();
  
      // Use font 1 in white.
      
      afOled.setTextSize(1);
      afOled.setTextColor(WHITE);

      // Create the title and trademark strings.
      
      String sTitle1("IntelliGrill");
      String sTitle2("Feather Server");
      String sTradeMark("TM");

      // Draw the trademark string.
      
      afOled.setCursor((DISPLAY_WIDTH / 2) + ((sTitle1.length() * FONT_ONE_WIDTH) / 2), 0);
      afOled.print(sTradeMark);

      // Draw the title string.
      
      afOled.setCursor((DISPLAY_WIDTH / 2) - ((sTitle1.length() * FONT_ONE_WIDTH) / 2), (DISPLAY_HEIGHT / 2) - (FONT_ONE_HEIGHT / 2));
      afOled.print(sTitle1);

      afOled.setCursor((DISPLAY_WIDTH / 2) - ((sTitle2.length() * FONT_ONE_WIDTH) / 2), DISPLAY_HEIGHT - FONT_ONE_HEIGHT);
      afOled.print(sTitle2);
  
      // Display the result.
      
      afOled.display();
      
  // Restore sSsid, sPassword and sPasswordIntelligrill from rtc ram.

      // rtcRead() will return true if successful, false if not.
      
      if(rtcRead())
      {
        // Read successful.
        
        Serial.println(String("Intelligrill data recovered from storage:"));
        Serial.println(String("  Intelligrill ssid            :  " + sSsidIntelligrill));
        Serial.println(String("  Wifi ssid                    :  " + sSsid));
        Serial.println(String("  Intelligrill & Wifi password :  " + sPassword + "\n"));
      }
      else
      {
        // Read failed.
        
        Serial.println(String("Intelligrill data recovery from storage FAILED.\n"));
        Serial.println(String("  Intelligrill ssid     :  " + sSsidIntelligrill));
        Serial.println(String("  Intelligrill password :  " + sPassword + "\n"));
      }
 
  // Wifi.

    // Set the wifi mode to access point and station mode.
    
    WiFi.mode(WIFI_AP_STA);

    // Set the station mode ssid and password.
    
    WiFi.begin(sSsid.c_str(), sPassword.c_str());

    // Set the access point ip address, gateway and subnet.
    
    WiFi.softAPConfig(IpAddressIntelligrill, IpGatewayIntelligrill, IpSubnetIntelligrill);

    // Set the access point ssid and password.
    
    WiFi.softAP(sSsidIntelligrill.c_str(), sPassword.c_str());

  // Buttons (the buttons are located on the Adafruit afOled module).
  
  pinMode(BUTTON_A, INPUT_PULLUP);
  pinMode(BUTTON_B, INPUT_PULLUP);
  pinMode(BUTTON_C, INPUT_PULLUP);
  
  // Temperature averaging.
  //
  // Fill the temperature averaging array nCounts[] with the current analog value.

  int nAnalogCurrent = analogRead(A0);
  for(int nCount = 0; nCount < COUNTS_ARRAY_SIZE; nCount ++)
    nCounts[nCount] = nAnalogCurrent;

  // Allow the title to display for a few seconds.

  delay(2000);

  // Then display the copyright.

      // Clear the display.

      afOled.clearDisplay();

      // Draw the copyright, date and holder.

      String sCopyrightLine("Copyright (c) 2017 by");
      afOled.setCursor((DISPLAY_WIDTH / 2) - ((sCopyrightLine.length() * FONT_ONE_WIDTH) / 2), 0);
      afOled.print(sCopyrightLine);

      sCopyrightLine = String("Zumwalt Properties,");
      afOled.setCursor((DISPLAY_WIDTH / 2) - ((sCopyrightLine.length() * FONT_ONE_WIDTH) / 2), (DISPLAY_HEIGHT / 2) - (FONT_ONE_HEIGHT / 2));
      afOled.print(sCopyrightLine);
      
      sCopyrightLine = String("LLC.");
      afOled.setCursor((DISPLAY_WIDTH / 2) - ((sCopyrightLine.length() * FONT_ONE_WIDTH) / 2), DISPLAY_HEIGHT - FONT_ONE_HEIGHT);
      afOled.print(sCopyrightLine);

      afOled.display();
      
      delay(2000);
      
     // Draw the rights.

      afOled.clearDisplay();
      
      sCopyrightLine = String("All Rights Reserved.");
      afOled.setCursor((DISPLAY_WIDTH / 2) - ((sCopyrightLine.length() * FONT_ONE_WIDTH) / 2), (DISPLAY_HEIGHT / 2) - (FONT_ONE_HEIGHT / 2));
      afOled.print(sCopyrightLine);

      afOled.display();

      delay(2000);  

  // By this time, wifi, if available, should be connected, start the server.

      // Root is the webpage.
      //
      // "/" is sent from the client and is a request for a complete webpage refresh.
      
      esp8266Server.on("/", []()
      {
        IntelligrillFeatherClient();
      });

      // /d is data.
      //
      // "/d" is sent from the client and is a request for a data update.
      
      esp8266Server.on("/d", []()
      {
        // Variables.
        
        String    sTimeRunning = DisplayFormatTime(dTimeRunning);
        String    sTimeToGo = DisplayFormatTime(dTimePredictedToGo);
  
        // Respond with current temperature, desired temperature, estimated time to go and time tunning to the client

        esp8266Server.send(200, "application/json", "{\"t\":\"" + String(nTemperatureCurrent) + "\", " +
                                                     "\"d\":\"" + String(nTemperatureDesired) + "\", " +
                                                     "\"g\":\"" + sTimeToGo + "\", " +  
                                                     "\"r\":\"" + sTimeRunning + "\"" +  
                                                     "}" );       
      });
      
      // /s is setting the desired temperature.
      //
      // "/s" is sent from the client and contains a new desired temperature.
      
      esp8266Server.on("/s", []()
      {
        // Update nTemperatureDesired with the client data.
        
        nTemperatureDesired = esp8266Server.arg("t").toInt();

        // Respond by echoing the value back to the client.

        esp8266Server.send(200, "application/json", "{\"d\":\"" + String(nTemperatureDesired) + "\"}");
      });

      // /setup is setting up intelligrill.
      
      esp8266Server.on("/setup", []()
      {
        // For increased security, only allow setup when the ip display page is shown on the oled.
        
        if(nDisplayMode == DISPLAY_MODE_IP)
        {
            // Intelligrill "IP Address" page is selected, allow setup.
    
            String    sSsidAccessPoint = esp8266Server.arg("ssidi");          
            String    sSsidWifi = esp8266Server.arg("ssid");   
                   
            if(sSsidAccessPoint.length() && sSsidAccessPoint.length())
            {
                // Have ssids, update ssids and password.
                
                sSsid = sSsidWifi;
                sSsidIntelligrill = sSsidAccessPoint;
                sPassword = esp8266Server.arg("password");
      
                // Reconnect with new ssid and password.
                
                WiFi.begin(sSsid.c_str(), sPassword.c_str());
                
                // Write the ssid and password to rtc memory.  rtcWrite() will return true if successfull,
                // false if not.
          
                if(rtcWrite())
                {
                    Serial.println(String("Intelligrill data written to storage:"));
                    Serial.println(String("  Intelligrill ssid            :  " + sSsidIntelligrill));
                    Serial.println(String("  Wifi ssid                    :  " + sSsid));
                    Serial.println(String("  Intelligrill & Wifi password :  " + sPassword + "\n"));
                }
            }
        
            // Refresh setup page on client.
            
            IntelligrillSetup();
        }
        else
        {
            // Intelligrill "IP Address" page is not selected, do not allow setup.

            IntelligrillSetupNotAllowed();
        }
      });

    // Start the server.
    
    esp8266Server.begin();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Main loop.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

void loop() 
{
   // Update server every pass.

   esp8266Server.handleClient();

  // Update either buttons, analog, Intelligrill or display every 10 passes.

    // Upcount nUpdate from 0 through 39.
    
    static  int     nUpdate = 0;
    nUpdate = (nUpdate == 39) ? 0 : nUpdate + 1;
  
    // Determine which, if any, get updated this pass.
    
    switch(nUpdate)
    {
      case 0:
      {
        // Update buttons.
    
        Buttons();
      }
      break;
  
      case 10:
      {
        // Update analog.
      
        Analog();
      }
      break;
  
      case 20:
      {
        // Upddate Intelligrill.
        
        Intelligrill();
      }
      break;
  
      case 30:
      {
        // Update display.
        
        Display();
      }
      break;
    }
    
  // Sleep.
  
  delay(DELAY_MILLISECONDS);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Analog
// Analog input averaging and temperature calculation.
// Entry   : nothing
// Returns : nothing
// Notes   :
//
//  The Maverick ET-72 thermistor temperator probe incorporated in this design is wired such that as
// temperature rises, the analog input falls and conversely as temperature falls, the analog input
// rises.  This method was chosen in order to allow the probe shield, which is internally connected
// to the connector ring, to be connected to ground for better noise reduction.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

void  Analog()
{
  // Add the current analog input count to nCounts[].
  
  nCounts[nCountsPointer ++] = analogRead(A0);
  nCountsPointer %= COUNTS_ARRAY_SIZE;

  // Calculate temperature.

    // Calculate the average of analog input counts.
    
    double dCountsAverage = 0.0;
    for(int nCount = 0; nCount < COUNTS_ARRAY_SIZE; nCount++)
        dCountsAverage += (double)nCounts[nCount];
    dCountsAverage /= (double)COUNTS_ARRAY_SIZE;

    // Adjust dCountsAverage polarity, offset and range.

    dCountsAverage = (1023.0 - (dCountsAverage + dCountsOffset)) * dCountsScale;
    
    // Limit dCountsAverage to 0 <= dCountsAverage <= COUNT_MAX.
        
    if(dCountsAverage < 0.0)
      dCountsAverage = 0.0;
    else if(dCountsAverage > AD_COUNT_MAX)
      dCountsAverage = AD_COUNT_MAX;

    // Look up the temperature.
    
    dTemperatureF = (nTemperatureTable[(int)(1023 - dCountsAverage)] / 10.0);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Buttons
// Update and process push buttons.
// Entry   : nothing
// Returns : nothing
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

void  Buttons()
{
  // Variables.
  
  static  int   nButtonA = 0;
  static  int   nButtonAN1 = 1;
  static  int   nButtonB = 0;
  static  int   nButtonBN1 = 1;
  static  int   nButtonC = 0;
  static  int   nButtonCN1 = 1;
  
  // Read buttons.
  
  int nDigitalReadA = digitalRead(BUTTON_A);
  int nDigitalReadB = digitalRead(BUTTON_B);
  int nDigitalReadC = digitalRead(BUTTON_C);
  
  // Process button A.

  if(nDigitalReadA != nButtonAN1)
  {
    // State changed.
    
    if(nButtonAN1)
    {
      // Pressed.
      
      nButtonAN1 = 0;
      nButtonA = 1;
      
      if(bEdit)
      {
        // Button A is currently being used to edit.
        
      }
      else
      {
        // Button A is currently being used to page.
        
        nDisplayMode = (nDisplayMode ? nDisplayMode - 1 : DISPLAY_MODE_COUNT - 1);
      }
    }
    else
    {
      // Released.
      
      nButtonAN1 = 1;
      nButtonA = 0;
    }
  }
  
  // If editing and button A is still down...

  if((bEdit) && (nDigitalReadC))
  {
    // Editing.
    
    if(nDigitalReadA)
    {
        // Button no longer down, zero speed.
        
        nEditIncrement = 0;
    }
    else
    {
      // Button still down, increase speed.

      if(nEditIncrement < 10)
        nEditIncrement ++;
    }
  }

   // Process button B.
  
  if(nDigitalReadB != nButtonBN1)
  {
    // State changed.
    
    if(nButtonBN1)
    {
      // Pressed.
      
      nButtonBN1 = 0;
      nButtonB = 1;
      
      // Toggle editing.
      
      if(bEdit)
      {
        // Stop editing.

        bEdit = false;
        nEditIncrement = 0;
      }
      else
      {
          // Start editing.

        if((nDisplayMode == DISPLAY_MODE_TIME_PREDICTED_TO_GO) || (nDisplayMode == DISPLAY_MODE_TIME_RUNNING))
        {
          // Just toggle seconds display mode.

          bSecondsDisplay = (bSecondsDisplay ? false : true);
        }
        else if(nDisplayMode == DISPLAY_MODE_TEMPERATURE_DESIRED)
        {
          // Start editing of desired temperature.
          
          bEdit = true;
          nEditIncrement = 0;
        }
      }
    }
    else
    {
      // Released.
      
      nButtonBN1 = 1;
      nButtonB = 0;
    }
  }
  
  // Check for "turn off".
  
  if(nButtonB)
  {
      // Button B is down, time it.
      
      nTurnOffDelay ++;
      
      if(nTurnOffDelay >= 20)
      {
          // Time is up, display good night.
        
          afOled.clearDisplay();
          String  sGoodNight("Good Night !");
          afOled.setTextSize(1);
          afOled.setCursor((DISPLAY_WIDTH / 2) - ((sGoodNight.length() * FONT_ONE_WIDTH) / 2), (DISPLAY_HEIGHT / 2) - (FONT_ONE_HEIGHT / 2));
          afOled.print(sGoodNight);
          afOled.display();
          delay(2000);

          // Clear display.
          
          afOled.clearDisplay();
          afOled.display();
          delay(500);

          // Go to sleep "forever".
          
          ESP.deepSleep(0);
          delay(1000);
      
      }
  }
  else
  {
      // Button B is up.
      
      nTurnOffDelay = 0;
  }

  // Process button C.
  
  if(nDigitalReadC != nButtonCN1)
  {
    // State changed.
    
    if(nButtonCN1)
    {
      // Pressed.
      
      nButtonCN1 = 0;
      nButtonC = 1;
      if(bEdit)
      {
        // Button C is currently being used to edit.
      
      }
      else
      {
        // Button C is currently being used to page.
      
        nDisplayMode = (nDisplayMode == DISPLAY_MODE_COUNT - 1 ? 0 : nDisplayMode + 1);
      }
    }
    else
    {
      // Released.
      
      nButtonCN1 = 1;
      nButtonC = 0;
    }
  }
    
  // If editing and button C is still down...

  if((bEdit) && (nDigitalReadA))
  {
    // Editing.
    
    if(nDigitalReadC == 0)
    {
      // Button still down, increase speed.
    
      if(nEditIncrement > -10)
        nEditIncrement --;
    }
    else
    {
        // Button no longer down, zero speed.
        
        nEditIncrement = 0;
    }
  }  
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Display
// Update afOled display.
// Entry   : nothing
// Returns : nothing
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// Display
// Update the oled display
// Entry   : nothing
// Returns : nothing
//

void  Display()
{  
  // Oled display.

  afOled.clearDisplay();

  // Decide which page to display.
  
  switch(nDisplayMode)
  {
    case  DISPLAY_MODE_IP:
    {
        // Display the ip address page.

                // Title.
                
                String  sTitle("IP Address");
                afOled.setTextSize(1);
                afOled.setCursor((DISPLAY_WIDTH / 2) - ((sTitle.length() * FONT_ONE_WIDTH) / 2), 0);
                afOled.print(sTitle);
                
                // Ip address.
                
                String  sIp(WiFi.localIP().toString());
                afOled.setCursor((DISPLAY_WIDTH / 2) - ((sIp.length() * FONT_ONE_WIDTH) / 2), DISPLAY_HEIGHT - FONT_ONE_HEIGHT);
                afOled.print(sIp);
  }
    break;

    case  DISPLAY_MODE_TEMPERATURE:
    {
        // Display the current temperature page.
        
            // Title.
            
            String  sTitle("Current Temperature");
            afOled.setTextSize(1);
            afOled.setCursor((DISPLAY_WIDTH / 2) - ((sTitle.length() * FONT_ONE_WIDTH) / 2), 0);
            afOled.print(sTitle);
            
            // Temperature.
            
            String  sTemperature(nTemperatureCurrent);
            afOled.setTextSize(2);
            afOled.setCursor((DISPLAY_WIDTH / 2) - ((sTemperature.length() * FONT_TWO_WIDTH) / 2), DISPLAY_HEIGHT - FONT_TWO_HEIGHT);
            afOled.print(sTemperature);
    }
    break;

    case  DISPLAY_MODE_TEMPERATURE_DESIRED:
    {
      // Display the desired temperature page.
    
            // Edit nTemperatureDesired.
            
            if(bEdit)
            {
              nTemperatureDesired += nEditIncrement;
              if(nTemperatureDesired < 100)
                nTemperatureDesired = 100;
              else if(nTemperatureDesired > 250)
                nTemperatureDesired = 250;
            }
            
            // Title.
            
            String  sTitle("Desired Temperature");
            afOled.setTextSize(1);
            afOled.setCursor((DISPLAY_WIDTH / 2) - ((sTitle.length() * FONT_ONE_WIDTH) / 2), 0);
            afOled.print(sTitle);
            
            // Temperature.
            
            String    sTemperature;
            if(bEdit)
              sTemperature = String("[ " + String(nTemperatureDesired) + " ]");
            else
              sTemperature = String(nTemperatureDesired);
            afOled.setTextSize(2);
            afOled.setCursor((DISPLAY_WIDTH / 2) - ((sTemperature.length() * FONT_TWO_WIDTH) / 2), DISPLAY_HEIGHT - FONT_TWO_HEIGHT);
            afOled.print(sTemperature);
    }
    break;

    case  DISPLAY_MODE_TIME_PREDICTED_TO_GO:
    {
      // Display time predicted to go page.
    
            // Title.
            
            String  sTemp("Estimated Time To Go");
            afOled.setTextSize(1);
            afOled.setCursor((DISPLAY_WIDTH / 2) - ((sTemp.length() * FONT_ONE_WIDTH) / 2), 0);
            afOled.print(sTemp);

            // Time.
            
            sTemp = DisplayFormatTime(dTimePredictedToGo);           
            afOled.setTextSize(2);
            afOled.setCursor(((DISPLAY_WIDTH / 2) - (sTemp.length() * FONT_TWO_WIDTH) / 2), DISPLAY_HEIGHT - FONT_TWO_HEIGHT);
            afOled.print(sTemp);
    }
    break;

    case  DISPLAY_MODE_TIME_RUNNING:
    {
        // Display time running page.
        
            // Title.
            
            String  sTemp("Running Time");
            afOled.setTextSize(1);
            afOled.setCursor((DISPLAY_WIDTH / 2) - ((sTemp.length() * FONT_ONE_WIDTH) / 2), 0);
            afOled.print(sTemp);
            
            // Time.
            
            sTemp = DisplayFormatTime(dTimeRunning);           
            afOled.setTextSize(2);
            afOled.setCursor(((DISPLAY_WIDTH / 2) - (sTemp.length() * FONT_TWO_WIDTH) / 2), DISPLAY_HEIGHT - FONT_TWO_HEIGHT);
            afOled.print(sTemp);
    }
    break;
 }

  // Present the display.
  
  afOled.display();
}

//
// DisplayFormatTime
// Format time for the display.
// Entry   : time
// Returns : string
//

String    DisplayFormatTime(double dTime)
{
  // Local variables.

  char    chBuffer[80];
  long    lTimeHours = (long)dTime / 3600;
  long    lTimeMinutes = (long)(dTime / 60) % 60;
  long    lTimeSeconds = (long)dTime % 60;

  // Generate formatted time.
  
  if(bSecondsDisplay)
  {
          if(dTime <= 0)
                  sprintf(chBuffer, "--:--:--");
          else
                  sprintf(chBuffer, "%02d:%02d:%02d", lTimeHours, lTimeMinutes, lTimeSeconds);
  }
  else
  {
          if(dTime <= 60)
                  sprintf(chBuffer, "--:--");
          else
                  sprintf(chBuffer, "%02d:%02d", lTimeHours, lTimeMinutes);
  }
 
  return  String(chBuffer);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Intelligrill
// Perform Intelligrill calculations.
// Entry   : nothing
// Returns : nothing
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

void  Intelligrill()
{
  // Update timing.

      // Obtain milliseconds since the board was powered on, account for wrap around.
      
      double  dMilliseconds = (double) millis();
      
      if(dMilliseconds < dMillisecondsLast)
        dMilliseconds += 2147483648.0;
          
  // Round double degrees f and convert to int.
  
      int nTemperatureF = (int)(dTemperatureF + 0.44444444444445);

  // Check for temperature change.

      if(nTemperatureF != nTemperatureCurrent)
      {
        // Temperature has changed, update nTemperatureLowest.
    
        if(nTemperatureF < nTemperatureLowest)
        {
          // Current temperature is lower than lowest temperature.
          // Set the highest and lowest temperatures equal to the
          // current temperature.
    
          nTemperatureHighest = nTemperatureLowest = nTemperatureF;
    
          // Then stop the time & temperature calculations.
    
          dMillisecondsRunning = 0.0;
          dTimePredicted = dTimePredictedToGo = -1.0;
          bRunning = false;
        }

      // Update nTemperatureHighest.
  
          if(nTemperatureF > nTemperatureHighest)
          {
            // Current temperature is highest temperature read.
      
            nTemperatureHighest = nTemperatureF;
          }

      // Monitor temperatures for stop and start.
  
          if(bRunning)
          {
            // Running, if the current temperature drops 10 degrees or more lower
            // than the highest temperature, stop the time and temperature
            // calculations.
      
            if((nTemperatureHighest - nTemperatureF) >= 10)
            {
              // Set the lowest temperature to the current temperature.
      
              nTemperatureLowest = nTemperatureF;
      
              // Stop the time and temperature calculations.
      
              dMillisecondsRunning = 0.0;
              dTimePredicted = dTimePredictedToGo = -1.0;
              bRunning = false;
            }
          }
          else
          {
            // Not running, if the current temperature is 5 or more degrees 
            // higher than the lowest temperature, then start the time and
            // temperature calculations.
      
            if((nTemperatureF - nTemperatureLowest) >= 5)
            {
              // Set the highest temperature to the current temperature.
              //
              // Note the highest temperature is set to the current temperature
              // in order to avoid a stop via the current temperature being 10
              // degrees or more lower than the highest temperture.
      
              nTemperatureHighest = nTemperatureF;
      
              // Start the time and temperature calcuations.
      
              bRunning = true;
            }
            else
            {
              // Still not running, reset milliseconds of running time.
      
              dMillisecondsRunning = 0.0;
            }
          }
    }

  // If still running, update running time.
  
    if(bRunning)
    {
      // Still running, update milliseconds of running time.
      
      dMillisecondsRunning += dMilliseconds - dMillisecondsLast;
    }
            
  // Update dMillisecondsLast.
  
      dMillisecondsLast = dMilliseconds;

  // Update predictions on temperature and/or time changes.

      if((nTemperatureCurrent != nTemperatureF) ||
         (dTimeRunning != (dMillisecondsRunning / 1000.0)))
      {
        // Update temperature.
        
        nTemperatureCurrent = nTemperatureF;

        // Update running time only while the current temperature is less than the desired temperature.

        if(nTemperatureCurrent < nTemperatureDesired)
        {
            dTimeRunning = dMillisecondsRunning / 1000.0;
            if(dTimeRunning > ((99.0 * 3600.0) + (59.0 * 60.0) + 59.0))
               dTimeRunning = ((99.0 * 3600.0) + (59.0 * 60.0) + 59.0);
            else if(dTimeRunning == 0)
              dTimeRunning = -1;
        }
        
        // Update predictions.
        
        Predict();
      }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Predict
// Predict dTimePredicted, dTimePredictedToGo.
// Entry   : nothing
// Returns : nothing
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

void  Predict()
{
  // If running, calculate dTimePredicted.
  
  if(bRunning)
  {
    // Running, check for finished.
    
    if(nTemperatureCurrent >= nTemperatureDesired)
    {
      // Finished.
      
      dTimePredicted = dTimeRunning;
      dTimePredictedToGo = 0;
    }
    else
    {
      // Not finished, compute dTimePredicted.
      
      if(nTemperatureDesired < nTemperatureCurrent)
      {
        // This should never happen, but just in case...
      
        nTemperatureDesired = nTemperatureCurrent;
      }
      
      // Validate the relationship between nTemperatureCurrent and nTemperatureLowest.
      
      if(nTemperatureCurrent <= nTemperatureLowest)
      {
        // nTemperatureCurrent <= nTemperatureLowest.  Note that nTemperatureCurrent cannot be lower
        // than nTemperatureLowest due to the computations for nTemperatureHighest and
        // nTemperatureLowest immediately after the computation for nTemperatureCurrent above.
        // However, nTemperatureCurrent could equal nTemperatureLowest if nTemperatureCurrent was
        // dropping.  If nTemperatureCurrent == nTemperatureLowest, a division by zero error
        // will occur in the dTimePredicted computation, thus a 1 is substituted in this case for
        // (nTemperatureCurrent - nTemperatureLowest).
        
        dTimePredicted = (dTimeRunning * (double)(nTemperatureDesired - nTemperatureLowest)) / 1.0;
      }
      else
      {
        // nTemperatureCurrent > nTemperatureLowest.  Compute dTimePredicted normally.
        
        dTimePredicted = (dTimeRunning * (double)(nTemperatureDesired - nTemperatureLowest)) /
                                         (double)(nTemperatureCurrent - nTemperatureLowest);
      }

      // The minimum lTimePredicted is 1 second while running.

      if(dTimePredicted <= 0)
        dTimePredicted = 1;

      // Calculate dTimePredictedToGo.
      
      dTimePredictedToGo = dTimePredicted - dTimeRunning;

      // The minimum dTimePredictedToGo is 1 second while running.

      if(dTimePredictedToGo <= 0)
        dTimePredictedToGo = 1;

      // Limit dTimePredictedToGo to 359,999 seconds (99:59:59).
      // This could occur if the user leaves the program active
      // with the probe reading a constant temperature for an
      // extended period of time.  Allowing the computation to
      // exceed this value is not necessary.

      if(dTimePredictedToGo > ((99.0 * 3600.0) + (59.0 * 60.0) + 59.0))
        dTimePredictedToGo = ((99.0 * 3600.0) + (59.0 * 60.0) + 59.0);
    }
  }

  // Not running.
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Real Time Clock Memory functions.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// rtcRamImageCrc
// Compute crc of current content of rtcRamImage.
// Entry   : nothing
// Returns : crc
// Notes   :
//
//  From the original Intelligrill socket transmissions.
//


uint32 rtcRamImageCrc()
{
  // Local variables.
  
  int      nCount = sizeof(rtcRamImage) - 4;            // This includes only ssid and password, not crc.
  uint32   nCrc = 0xffffffff;                           // Initial crc pattern.
  uint8*   pRtcRamImage = (uint8 *) & rtcRamImage;      // "Pointer" to the start of the rtcRamImage.

  // Cacluate crc.
  
  while(nCount--)
  {
    // Read the next character.
    
    uint8 cChar = * pRtcRamImage ++;

    // Update crc for this character.
    
    for(uint32 i = 0x80; i > 0; i >>= 1) 
    {
      bool bBit = nCrc & 0x80000000;
      
      if(cChar & i) 
      {
        bBit = ! bBit;
      }
      
      nCrc <<= 1;
      
      if(bBit) 
      {
        nCrc ^= 0x04c11db7;
      }
    }
  }

  // Return.
  
  return nCrc;
}

//
// rtcRead
// Read sSsid and sPassword from rtc memory.
// Entry   : nothing
// Returns : true if successful, false if not
//

bool      rtcRead()
{
    // Read rtc memory into rtcRamInage.

    if(system_rtc_mem_read(RTC_MEMORY_START, & rtcRamImage, sizeof(rtcRamImage)))
    {
        // Read successful, check stored crc agains rtcRamImage crc

        if(rtcRamImage.nCrc == rtcRamImageCrc())
        {
            // crc match, restore sSsids 
        
            sSsid = String((char *) &rtcRamImage.chSsid);
            sSsidIntelligrill = String((char *) &rtcRamImage.chSsidAp);
            
            // Restore password.

            sPassword = String((char *) &rtcRamImage.chPassword);

            // Read succeeded.

            return true;
        }
        else
        {
            // Read succeeded, yet crc failed.

            Serial.println(String("Intelligrill data recovered from storage:"));
            Serial.println(String(" rtcRead :  crc fail.\n"));
            return false;
        }
    }
    else
    {
        // Read failed.

        Serial.println(String("Intelligrill data recovered from storage:"));
        Serial.println(String("  rtcRead :  rtc memory read fail.\n"));
        return false;
    }
}

//
// rtcWrite
// Write sSsid and sPassword to rtc memory.
// Entry   : nothing
// Returns : true if successful, false if not
//

bool      rtcWrite()
{
    // Intialize rtcRamImage.

        // Clear rtcRamImage.
        
        memset(& rtcRamImage, sizeof(rtcRamImage), 0);

        // Copy sSsids to rtcRamImage
        
        sSsid.toCharArray((char *)rtcRamImage.chSsid, sSsid.length() + 1);
        sSsidIntelligrill.toCharArray((char *)rtcRamImage.chSsidAp, sSsidIntelligrill.length() + 1);
        
        // Copy sPassword to rtcRamImage.chPassword.
        
        sPassword.toCharArray((char *)rtcRamImage.chPassword, sPassword.length() + 1);

        // Calculate and copy crc of rtcRamImage to rtcRamImage.nCrc.

        rtcRamImage.nCrc = rtcRamImageCrc();
    
    // Write rtcRamImage to rtc ram.

    return  system_rtc_mem_write(RTC_MEMORY_START, & rtcRamImage, sizeof(rtcRamImage));
}

// Copyright.

  const char*               chCopyrightFinal =
                            "Intelligrill (TM) Feather Copyright 2017 by Zumwalt Properties, LLC.  All Rights Reserved";

